__author__ = 'thorwhalen'

from datapath import datapath
import pickle
import os
import pandas as pd
import pfile.to as file_to
import pfile.name as file_name
import pstr.to as str_to
from khan_utils.encoding import to_unicode_or_bust

class DataAccessor(object):

    def __init__(self, data_root_location=None, extension='', force_extension = False, encoding='UTF-8'):
        self.root_folder = datapath(data_root_location)
        self.extension = extension
        self.force_extension = force_extension
        self.encoding = encoding

    def filepath(self,filename):
        # remove slash suffix if present
        if filename.startswith('/'):
            filename = filename[1:]
        # add extension
        if self.extension:
            if self.force_extension:
                if filename=='':
                    filename = file_name.replace_extension(filename)
                else:
                    filename = file_name.replace_extension(filename,self.extension)
            else:
                if filename=='':
                    filename = file_name.add_extension_if_not_present(filename)
                else:
                    filename = file_name.add_extension_if_not_present(filename,self.extension)
        return os.path.join(self.root_folder,filename)


    def dumpo(self, obj, filename):
        """
        saves an object to a local location
        """
        pickle.dump(obj,open(self.filepath(filename),'w'))

    def loado(self, filename):
        """
        loads an object from a local location
        """
        return pickle.load(open(self.filepath(filename),'r'))


    def dumps(self, the_str, filename, encoding="UTF-8"):
        """
        saves an object to a local location
        """
        str_to.file(the_str, self.filepath(filename), encoding=encoding)

    def loads(self, filename):
        """
        loads an object from a local location
        """
        return file_to.string(self.filepath(filename))

    def dumpu(self, the_str, filename, encoding="UTF-8"):
        """
        saves an object to a local location
        """
        str_to.file(to_unicode_or_bust(the_str), self.filepath(filename), encoding=encoding)

    def loadu(self, filename):
        """
        loads an object from a local location
        """
        return to_unicode_or_bust(file_to.string(self.filepath(filename)))


    def df_to_csv(self, df, filename):
        """
        saves an object to a local location
        """
        return df.to_csv(self.filepath(filename),encoding='UTF-8',sep="\t")

    def df_to_excel(self, df, filename):
        """
        saves an object to a local location
        """
        if filename.startswith('/'):
            filename = filename[1:]
        return df.to_excel(self.filepath(filename))


__author__ = 'thorwhalen'

from ut.datapath import datapath
import pickle
import os
from nippy.importing import get_environment_variable
import pandas as pd
import epysode.to as file_to
import epysode.name as pfile_name
import typyst.to as str_to
from typyst.trans import str_to_unicode_or_bust
#from os import environ # does this load the whole array? Can we just take MS_DATA instead?

try:
    MS_DATA = get_environment_variable('MS_DATA')
except KeyError:
    MS_DATA = ''

ENCODING_NONE = 0
ENCODING_UNICODE = 1
ENCODING_UTF8 = 2


class Local(object):

    def __init__(self, relative_root=None, extension='', force_extension=False, encoding='UTF-8', mother_root=MS_DATA):
        if relative_root:
            self.root_folder = os.path.join(mother_root, relative_root)
        else:
            self.root_folder = mother_root
        self.extension = extension
        self.force_extension = force_extension
        self.encoding = encoding


    def filepath(self, filename=None):

        if not filename:
            return self.root_folder

        # remove slash suffix if present (because self.sound_file_root_folder ends with / already)
        if filename.startswith('/'):
            filename = filename[1:]
        # add extension
        if self.extension:
            if self.force_extension:
                filename = pfile_name.replace_extension(filename, self.extension)
            else:
                filename = pfile_name.add_extension_if_not_present(filename, self.extension)
        return os.path.join(self.root_folder, filename)


    def dumpo(self, obj, filename):
        """
        saves an object to a local location
        """
        pickle.dump(obj,open(self.filepath(filename),'w'))

    def loado(self, filename):
        """
        loads an object from a local location
        """
        return pickle.load(open(self.filepath(filename),'r'))


    def dumps(self, the_str, filename, encoding=None):
        """
        saves an object to a local location
        """
        encoding = encoding or self.encoding
        str_to.file(the_str, self.filepath(filename), encoding=encoding)


    def loads(self, filename):
        """
        loads an object from a local location
        """
        return file_to.string(self.filepath(filename))


    def dumpu(self, the_str, filename, encoding=None):
        """
        saves an object to a local location
        """
        encoding = encoding or self.encoding
        str_to.file(str_to_unicode_or_bust(the_str), self.filepath(filename), encoding=encoding)


    def loadu(self, filename):
        """
        loads an object from a local location
        """
        return str_to_unicode_or_bust(file_to.string(self.filepath(filename)))


    def df_to_csv(self, df, filename, encoding=None):
        """
        saves an object to a local location
        """
        encoding = encoding or self.encoding
        return df.to_csv(self.filepath(filename), encoding=encoding, sep="\t")


    def df_to_excel(self, df, filename, **kwargs):
        """
        saves an object to a local location
        """
        if filename.startswith('/'):
            filename = filename[1:]
        if 'index' not in kwargs: # make the default index=False if not specified in input
            kwargs = dict({'index': False}, **kwargs)
        return df.to_excel(self.filepath(filename), **kwargs)

    def excel_to_df(self, filename, sheetname=None, **kwargs):
        """
        returns a df from the excel file specified by filename, taking the sheet specified
        by sheetname (which should be the sheetname, but could also be the sheet number (starting by 0)). Default
        sheetname is the first sheet.
        Further kwargs can be passed to pd.ExcelFile.parse
        """
        xd = pd.ExcelFile(self.filepath(filename))
        if sheetname is None:
            sheetname = xd.sheet_names[0]
            print("No sheetname specified so I'm taking the first one: %s" % sheetname)
        elif isinstance(sheetname, int):
            sheetname = xd.sheet_names[sheetname]
        return xd.parse(sheetname, **kwargs)


__author__ = 'thorwhalen'


import epysode.name as pfile_name
# from ut.khan_utils.encoding import to_unicode_or_bust
from pygeonhole.local import Local
from pygeonhole.s3 import S3


class DataAccessor(object):

    LOCATION_LOCAL = 'LOCAL'
    LOCATION_S3 = 'S3'

    def __init__(self, relative_root=None, mother_root=None, extension=None, force_extension=False, encoding='UTF-8', location=LOCATION_LOCAL, **kwargs):

        # remember these so can switch between locations without having to reset
        self.relative_root = relative_root
        self.extension = extension or ''
        self.force_extension = force_extension or False
        self.encoding = encoding
        self.mother_root = mother_root

        if location==self.LOCATION_LOCAL:
            self.use_local(**kwargs)
        elif location==self.LOCATION_S3:
            self.use_s3(**kwargs)
        else:
            raise AttributeError("Don't know that location (use 'LOCAL' or 'S3')")

    def use_local(self, relative_root=None, mother_root=None, extension=None, force_extension=None, encoding='UTF-8', **kwargs):

        self.currently_using = self.LOCATION_LOCAL

        # replace self vars with any new vals passed in by client
        self.relative_root = relative_root or self.relative_root
        self.extension = extension or self.extension
        self.force_extension = force_extension or self.force_extension
        self.encoding = encoding or 'UTF-8'

        # TODO - we need to update local to accept mother_root argument
        self.local = Local(relative_root=self.relative_root, extension=self.extension, force_extension=self.force_extension, encoding=self.encoding, **kwargs)
        self.dacc = self.local

    def use_s3(self, relative_root=None, mother_root=None, extension=None, force_extension=None, encoding=None, **kwargs):

        self.currently_using = self.LOCATION_S3

        # replace self vars with any new vals passed in by client
        self.relative_root = relative_root or self.relative_root
        self.extension = extension or self.extension
        self.force_extension = force_extension or self.force_extension
        self.encoding = encoding or self.encoding
        self.mother_root = mother_root or self.mother_root
        if self.mother_root is None:
            if self.relative_root.startswith('/'):
                self.relative_root = self.relative_root[1:]
            self.mother_root = pfile_name.get_highest_level_folder(self.relative_root)
            self.relative_root = self.relative_root.replace(self.mother_root,'',1)

        # set folder and other props on s3 via constructor or other
        self.s3 = S3(self.mother_root, self.relative_root, self.extension, self.force_extension, self.encoding, **kwargs)
        self.dacc = self.s3


    def load_excel(self, filename):
        data = self.dacc.loado


    ####################################################################################################################

    def __getattr__(self, name):

        def _missing(*args, **kwargs):

            if self.currently_using == self.LOCATION_LOCAL:
                target = self.local
            elif self.currently_using == self.LOCATION_S3:
                target = self.s3
            else:
                raise RuntimeError("Need to implement method missing for {}.".format(self.currently_using))

            if name in dir(target):
                return getattr(target, name)(*args, **kwargs)
            else:
                raise AttributeError("{} does not exist in {}".format(name, target))

        return _missing



    #
    # def dumpo(self, obj, filename):
    #     """
    #     """
    #
    #     self.dacc.dumpo(obj, filename)
    #     # pickle.dump(obj,open(self.filepath(filename),'w'))
    #
    # def loado(self, filename):
    #     """
    #     loads an object from a local location
    #     """
    #     self.dacc.loado(filename)
    #
    #
    # def dumps(self, the_str, filename, encoding=None):
    #     """
    #     """
    #     if encoding:
    #         s = encoding(the_str)
    #
    #     self.
    #
    #     if self.location == 'local':
    #         self.dacc.dumps(the_str, filename, encoding=encoding)
    #     elif self.location == 's3':
    #         self.dacc.dumps(the_str, filename)
    #
    #
    # def loads(self, filename):
    #     """
    #     loads an object from a local location
    #     """
    #     self.dacc.loads(filename)
    #
    #
    # def dumpu(self, the_str, filename, encoding=None):
    #     """
    #     saves an object to a local location
    #     """
    #     self.dacc.dumps(to_unicode_or_bust(the_str), filename, encoding=encoding)
    #
    #
    # def loadu(self, filename):
    #     """
    #     loads an object from a local location
    #     """
    #     return to_unicode_or_bust(self.dacc.loads(self.filepath(filename)))
    #
    #
    # def df_to_csv(self, df, filename, encoding=None):
    #     """
    #     saves an object to a local location
    #     """
    #     encoding = encoding or self.encoding
    #     return df.to_csv(self.filepath(filename), encoding=encoding, sep="\t")
    #
    #
    # def df_to_excel(self, df, filename):
    #     """
    #     saves an object to a local location
    #     """
    #     if filename.startswith('/'):
    #         filename = filename[1:]
    #     return df.to_excel(self.filepath(filename))





__author__ = 'mattjmorris'

from .dynamo import Dynamo
from boto.dynamodb2.table import Table
from datetime import datetime
from khan_utils.encoding import to_unicode_or_bust, to_utf8_or_bust
from ut.coll import order_conserving


class DDBSlurps(Dynamo):

    @classmethod
    def from_test_mode(cls, access_key=None, secret=None):
        """
        Use this for getting an instance of this class that uses test tables.
        """
        instance = cls(access_key, secret)
        instance.slurps_table = Table('test_slurps', connection=instance.connection)
        instance.failed_slurps_table = Table('test_failed_slurps', connection=instance.connection)
        return instance

    def __init__(self, access_key=None, secret=None):
        """
        ! Use test_mode factory method for instantiating this class with test_slurps and test_failed_slurps tables
        """
        super(DDBSlurps, self).__init__(access_key, secret)

        self.slurps_table = Table('slurps', connection=self.connection)
        self.failed_slurps_table = Table('failed_slurps', connection=self.connection)

    def save_slurp_info(self, slurp_info_, overwrite=True):
        """
        slurp_info_ can either be in the form of a list of dicts or else a single dict.
        If slurp_info is a list, batch write will be used
        """
        if isinstance(slurp_info_, dict):
            self.slurps_table.put_item(slurp_info_, overwrite=overwrite)
        elif isinstance(slurp_info_, list):
            with self.slurps_table.batch_write() as batch:
                for s in slurp_info_:
                    batch.put_item(data=s, overwrite=overwrite)
        else:
            raise TypeError("slurp_info must be a dict or a list of dicts, not a {}".format(type(slurp_info_)))

    def save_failed_slurp(self, searchterm):
        self.failed_slurps_table.put_item(data={'searchterm': searchterm, 'datetime': datetime.now().isoformat()},
                                          overwrite=True)

    def get_slurp_info(self, search_term_=None):
        """
        search_term_ can be either a string or a list of strings. Each string should be a search term you are looking
        for in the db.
        Returns either a single list of key-value tuples (if search_term_ was a string)
        or a list of key-value tuples (if search_term_ was a list)
        Each list of key-value tuples can easily be converted to a dict or an OrderedDict by the client.
        """

        # searchterm_ is a STRING
        if isinstance(search_term_, str):
            if search_term_:
                slurp_info = list((self.slurps_table.get_item(searchterm=search_term_)).items())
            else:
                slurp_info = []

        # searchterm is a LIST of strings
        elif isinstance(search_term_, list):
            if search_term_:
                # create a set of non-empty searchterms. We us a set to avoid a duplicate query error from the db
                set_of_sts = {st for st in search_term_ if st}
                # create a list of dicts from the set
                list_of_st_dicts = [{'searchterm': st} for st in set_of_sts]
                res = self.slurps_table.batch_get(list_of_st_dicts)
                try:
                    slurp_info = [list(i.items()) for i in res]
                except (StopIteration, IndexError):
                    # If res is empty, we get one of these errors when trying to iterate.
                    slurp_info = []
            else:
                slurp_info = []

        # searchterm is an unexpected type
        else:
            raise TypeError("search_term_ must be a dict or a list of dicts, not a {}".format(type(search_term_)))

        return slurp_info

    def existing_and_missing_uni(self, searchterm_list):
        """
        Takes a list of searchterm strings and returns a list of searchterm strings that were found in the db (in unicode)
        and a list of the searchterms that were missing from the found results
        """
        # make sure in utf8 before we send request to the db
        input_sts_utf8 = [to_utf8_or_bust(i) for i in searchterm_list]
        found_sts_info = self.get_slurp_info(input_sts_utf8)
        found_sts_uni = [to_unicode_or_bust(dict(i)['searchterm']) for i in found_sts_info]
        input_sts_uni = [to_unicode_or_bust(i) for i in input_sts_utf8]
        missing_sts_uni = order_conserving.setdiff(input_sts_uni, found_sts_uni)
        return found_sts_uni, missing_sts_uni

    def get_table(self, table_name):
        """
        Convenience method for client who may wish to get a specific table from the DynamoDB connection
        """
        return Table(table_name, connection=self.connection)

    def truncate_failed_slurp_table(self):
        """
        """
        with self.failed_slurps_table.batch_write() as batch:
            for item in self.failed_slurps_table.scan():
                batch.delete_item(searchterm=item['searchterm'])

    def truncate_slurp_table(self):
        """
        WARNING! Only use for test mode table
        """
        assert self.slurps_table.table_name == 'test_slurps', "Will only truncate test slurps table. To truncate production table, run code manually"
        test_slurps_table = Table('test_slurps', connection=self.connection)
        with test_slurps_table.batch_write() as batch:
            for item in self.slurps_table.scan():
                batch.delete_item(searchterm=item['searchterm'])

    def modify_failed_slurps_throughput(self, requested_read, requested_write):
        return self.modify_throughput(requested_read, requested_write, self.failed_slurps_table)

    def modify_slurps_throughput(self, requested_read, requested_write):
        return self.modify_throughput(requested_read, requested_write, self.slurps_table)

    def get_slurps_table_info(self):
        return self.get_table_info(self.slurps_table)

    def get_failed_slurps_table_info(self):
        return self.get_table_info(self.failed_slurps_table)




# Author: Tom Burge
# Contributor: Anthon van der Neut - Thanks for the help!
# Title: py-yaml-json-parser
# Description: Simple Menu-Based YAML/JSON Parser
# Requirements:
#   - Ruamel.YAML (pip install ruamel.yaml
#
# See: https://stackoverflow.com/questions/51914505/python-yaml-to-json-to-yaml for explanations

import json
import sys
from collections.abc import Mapping, Sequence
from collections import OrderedDict
import ruamel.yaml
from ruamel.yaml.error import YAMLError
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.scalarstring import PreservedScalarString, SingleQuotedScalarString
from ruamel.yaml.compat import string_types, MutableMapping, MutableSequence

yaml = ruamel.yaml.YAML()


class OrderlyJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Mapping):
            return OrderedDict(o)
        elif isinstance(o, Sequence):
            return list(o)
        return json.JSONEncoder.default(self, o)


def printmenu():
    menu = ('1. Parse YAML to JSON\n'
            '2. Parse JSON to YAML\n'
            '3. Exit\n'
            )
    print(menu)


def preserve_literal(s):
    return PreservedScalarString(s.replace('\r\n', '\n').replace('\r', '\n'))


def walk_tree(base):
    if isinstance(base, MutableMapping):
        for k in base:
            v = base[k]  # type: Text
            if isinstance(v, string_types):
                if '\n' in v:
                    base[k] = preserve_literal(v)
                elif '${' in v or ':' in v:
                    base[k] = SingleQuotedScalarString(v)
            else:
                walk_tree(v)
    elif isinstance(base, MutableSequence):
        for idx, elem in enumerate(base):
            if isinstance(elem, string_types):
                if '\n' in elem:
                    base[idx] = preserve_literal(elem)
                elif '${' in elem or ':' in elem:
                    base[idx] = SingleQuotedScalarString(elem)
            else:
                walk_tree(elem)

def parseyaml(intype, outtype):
    infile = input('Please enter a {} filename to parse: '.format(intype))
    outfile = input('Please enter a {} filename to output: '.format(outtype))

    with open(infile, 'r') as stream:
        try:
            datamap = yaml.load(stream)
            with open(outfile, 'w') as output:
                output.write(OrderlyJSONEncoder(indent=2).encode(datamap))
        except YAMLError as exc:
            print(exc)
            return False
    print('Your file has been parsed.\n\n')


def parsejson(intype, outtype):
    infile = input('Please enter a {} filename to parse: '.format(intype))
    outfile = input('Please enter a {} filename to output: '.format(outtype))

    with open(infile, 'r') as stream:
        try:
            datamap = json.load(stream, object_pairs_hook=CommentedMap)
            walk_tree(datamap)
            with open(outfile, 'w') as output:
                yaml.dump(datamap, output)
        except YAMLError as exc:
            print(exc)
            return False
    print('Your file has been parsed.\n\n')


loop = True

while loop:
    printmenu()  # Prints Menu for User
    choice = int(input('Please select an operation: '))

    if choice == 1:
        infiletype = 'YAML'
        outfiletype = 'JSON'
        parseyaml(infiletype, outfiletype)
    elif choice == 2:
        infiletype = 'JSON'
        outfiletype = 'YAML'
        parsejson(infiletype, outfiletype)
    elif choice == 3:
        sys.exit(0)
    else:
        print('Please make a selection')





__author__ = 'mattjmorris'

from .dynamo import Dynamo
from boto.dynamodb2.table import Table

class DDBKids(Dynamo):

    @classmethod
    def from_test_mode(cls, access_key=None, secret=None):
        """
        Use this for getting an instance of this class that uses test tables.
        """
        instance = cls(access_key, secret)
        instance.table = Table('test_kids', connection=instance.connection)
        return instance

    def __init__(self, access_key=None, secret=None):
        super(DDBKids, self).__init__(access_key, secret)
        self.table = Table('kids', connection=self.connection)

    def set_max_kid(self, account_name, kid):
        """
        Set the max kid used for an account
        :param account_name:
        :param kid: int >= 0.
        """
        return self.table.put_item({'account_name': account_name, 'kid': kid}, overwrite=True)

    def get_max_kid(self, account_name):
        """
        Get the max kid already used for an account. If the account does not exist, create it in the db with a KID of 0.
        :param account_name:
        """
        res = self.table.get_item(account_name=account_name)
        if res['kid']:
            return int(res['kid'])
        else:
            self.logger.warn("Creating a new (max) KID entry for account {} because it did not yet exist in ddb_kids".format(account_name))
            self.set_max_kid(account_name, 0)
            return 0

    def modify_throughput(self, requested_read, requested_write, table=None):
        table = table or self.table
        return super(DDBKids, self).modify_throughput(requested_read, requested_write, table)

    def truncate_table(self):
        """
        WARNING! Only use for test mode table
        """
        assert self.table.table_name == 'test_kids', "Will only truncate test table. To truncate production table, run code manually"
        with self.table.batch_write() as batch:
            for item in self.table.scan():
                batch.delete_item(dt=item['dt'])


__author__ = 'thorwhalen'

import logging
import traceback
import json
from datetime import datetime
import pandas as pd
from pandas import DataFrame
import os
from nippy.importing import get_environment_variable

import numpy as np
from serialize.amazon_sender import AmazonSender
from collections import OrderedDict


class Logger(object):

    @classmethod
    def get_most_recent_log_as_df(cls):
        return cls.get_log_as_df(cls.most_recent_log())

    @classmethod
    def get_most_recent_error(cls):
        return cls.get_last_error(cls.most_recent_log())

    @classmethod
    def get_log_as_df(cls, file_name_and_path):

        with open(file_name_and_path) as f:
            content = f.readlines()

        # only pull in rows that comply with json
        data = [json.loads(d) for d in content if d.startswith('{')]

        if data:
            df = DataFrame(data)
            df['dt'] = pd.to_datetime(df['dt'])
            df = df.set_index('dt')
            # Get columns we are interested in to show up in order we want them in
            cols_of_interest = [col for col in ['msg', 'origin', 'level', 'error_traceback'] if col in df.columns]
            cols_not_of_interest = list(set(df.columns) - set(cols_of_interest))
            df = df[cols_of_interest + cols_not_of_interest]
        else:
            df = DataFrame()

        return df.fillna('')

    @classmethod
    def get_last_error(cls, file_name_and_path):
        df = cls.get_log_as_df(file_name_and_path)
        if df:
            df = df[df['error_traceback'].notnull()]
            df.sort_index(ascending=False, inplace=True)
            return df.iloc[0]['error_traceback']
        else:
            return ''

    @classmethod
    def most_recent_log(cls, base_folder=None):
        logdir = base_folder or get_environment_variable('DATA_LOG_FOLDER')
        logfiles = sorted([f for f in os.listdir(logdir) if f.endswith('.log')])
        most_recent_log = os.path.join(logdir, logfiles[-1])
        return most_recent_log

    @classmethod
    def move_logs_to_backup(cls, base_folder=None):

        logdir = base_folder or get_environment_variable('DATA_LOG_FOLDER')
        backup_folder = os.path.join(logdir, 'bak')
        if not os.path.exists(backup_folder):
            os.makedirs(backup_folder)
        # delete out old ones so only most recent old one is ever there
        else:
            for f in os.listdir(backup_folder):
                file_path = os.path.join(backup_folder, f)
                try:
                    if os.path.isfile(file_path):
                        os.unlink(file_path)
                except Exception as e:
                    print(e)

        for the_file in os.listdir(logdir):
            try:
                if the_file.endswith(".log"):
                    os.rename(os.path.join(logdir, the_file), os.path.join(backup_folder, the_file))
            except Exception as e:
                print(e)

    def __init__(self, level=logging.INFO, file_name='default_log', make_file_name_unique=False, mode='a', origin=''):
        """
        Main purpose of this logger is to make sure everything gets logged in JSON format, so that it can easily
        be read back in the form of a Pandas dataframe.

        Mode: w for write, a for append. Defaults to w.
        """

        if file_name:
            if os.path.exists(file_name) and os.path.exists(file_name):
                name_and_path = file_name
            else:
                name_and_path = os.path.join(get_environment_variable('DATA_LOG_FOLDER'), file_name)

        self.filename_and_path = name_and_path

        generic_folder = os.path.join(get_environment_variable('DATA_LOG_FOLDER'), 'generic')
        if not os.path.exists(generic_folder):
            os.makedirs(generic_folder)
        generic_log_name_and_path = os.path.join(generic_folder, "generic.log")
        logging.basicConfig(filemode='w', filename=generic_log_name_and_path, level=level, format='%(message)s')

        # This next line uses the name_and_path as a unique namespace
        self.logger = logging.getLogger(name_and_path)
        #self.logger.setLevel(level)
        #self.logger
        self.logger.handlers = []
        fh = logging.FileHandler(name_and_path, mode='a')
        fh.setLevel(level)
        formatter = logging.Formatter('%(message)s')
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)
        #self.logger

        self.origin = origin

        self.email = AmazonSender()
        self.num_errors = 0

    def log(self, level=logging.DEBUG, msg=None, origin=None, error_obj=None, **kwargs):

        origin = origin or self.origin

        # construct a dict to log
        log_dict = OrderedDict()

        if msg:
            log_dict['msg'] = msg
        log_dict['origin'] = origin
        log_dict['level'] = logging.getLevelName(level)
        log_dict['dt'] = datetime.now().isoformat()
        log_dict.update(kwargs)

        # MJM: the traceback object holds on to errors, so only ask it for info if we are logging an error
        if level == logging.ERROR:
            # If there was an exception, let's grab it here
            tb = traceback.format_exc()
            if tb and tb!="None\n":
                log_dict['error_traceback'] = tb
        elif error_obj:
            log_dict['error_traceback'] = str(error_obj)

        self.logger.log(level=level, msg=json.dumps(log_dict))

    def debug(self, msg, **kwargs):
        self.log(level=logging.DEBUG, msg=msg, **kwargs)

    def info(self, msg, **kwargs):
        self.log(level=logging.INFO, msg=msg, **kwargs)

    def warn(self, msg, error_obj=None, **kwargs):
        self.log(level=logging.WARN, msg=msg, error_obj=error_obj, **kwargs)

    def error(self, msg, error_obj=None, send_email=True, **kwargs):
        """
        For message, it is recommended you pass in the message from the error_obj object (if there is one), or else a
        custom message that includes the error_obj message.
        The traceback will be automatically extracted from the traceback object.
        If sending email, it is recommended you pass in the original error_obj object
        """
        self.log(level=logging.ERROR, msg=msg, error_obj=error_obj, **kwargs)
        if send_email:
            body_html = "<p> Last log entries: </p>"
            body_html += "</br>"

            full_error = None

            try:
                df = self.get_current_log_as_df()
                body_html += df[-10:].to_html()
                self.num_errors += 1

                try:
                    df_sorted = df.sort_index(ascending=False)
                    full_error = df_sorted.iloc[0]['error_traceback']
                except:
                    pass
            except Exception as e:
                self.warn("Could not generate error df", error_obj=e)

            error_html = full_error or str(error_obj) or msg
            # force spacing
            error_html = error_html.replace(' ', '&nbsp;')
            # put brs where there had been newlines
            error_html = error_html.replace('\n', '<br />')
            error_html = "<p>" + error_html + "</p>"
            error_html += body_html

            if self.num_errors == 10:
                self.email.send_email(subject='10th KHAN Error, last one sending', text="Error", html=error_html)
            elif self.num_errors < 10:
                self.email.send_email(subject='KHAN Error', text="Error", html=error_html)

    # DEPRECATED
    #def structured_log(self, level=logging.DEBUG, origin=None, msg='', **kwargs):
    #    origin = origin or self.origin or ''
    #    self.log(level=level, origin=origin, msg=msg, **kwargs)

    # DEPRECATED
    #def id_and_reason_log(self, id, reason, level=logging.ERROR):
    #    self.log(level=level, id=id, reason=reason)

    def get_current_log_as_df(self):
        """
        Calls the class-level method for getting df of of a file, passing in path of current file
        """
        return self.get_log_as_df(self.filename_and_path)

    def get_current_error(self):
        """
        Calls the class-level method for getting last error from a file, passing in path of current file
        """
        return self.get_last_error(self.filename_and_path)

    def _make_file_name_and_path(self, file_name, file_path, make_file_name_unique):

        file_path = file_path or get_environment_variable('DATA_LOG_FOLDER')

        if make_file_name_unique:
            time_str = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            name = os.path.splitext(file_name)[0]
            extension = os.path.splitext(file_name)[1]
            if name:
                name += "_"
            file_name = name + time_str + extension

        if not os.path.splitext(file_name)[1]:
            file_name += '.log'

        full_name_and_path = os.path.join(file_path, file_name)

        return full_name_and_path

    # DEPRECATED
    def llog(self, level=None, **kwargs):
        """
        exactly like log, except it defaults to self.logger.level instead of to logging.DEBUG
        (llog for Level-LOG)
        """

        # default to object's logger.level
        level = level or self.logger.level

        # call log
        self.log(level=level, **kwargs)

    # DEPRECATED
    @classmethod
    def get_reason_stats(cls, log_df):
        nRows = len(log_df)
        stats_df = log_df[['reason']].dropna()
        nNanRows = nRows - len(stats_df)
        stats_df = stats_df.groupby(['reason']).count()
        stats_df = stats_df.rename(columns={'reason': 'count'})
        #stats_df = ch_col_names(stats_df, ['count'], ['reason'])
        #return stats_df
        stats_df = pd.concat([stats_df,
                              pd.DataFrame({'reason': ['NAN'], 'count': [nNanRows]}).set_index('reason')])
        #return stats_df.order(ascending=False)
        return stats_df.sort(columns=['count'], ascending=False)

    # DEPRECATED
    @classmethod
    def get_no_error_rows(cls, log_df):
        return log_df[~np.isnan(log_df.reason)]


import os
from time import sleep
import boto.dynamodb2
from .khan_logger import KhanLogger

__author__ = 'mattjmorris'

class Dynamo(object):

    def __init__(self, access_key=None, secret=None):
        """
        If access_key and/or secret are not passed in, assumes we are accessing erenev's aws account and that the
        access info is stored as environment variables on the current server.

        Connection and Table are available to clients via self properties, in case clients wish to use those objects
        directly.
        """
        access_key = access_key or os.getenv('VEN_S3_ACCESS_KEY')
        secret = secret or os.getenv('VEN_S3_SECRET')
        self.connection=boto.dynamodb2.connect_to_region(region_name='eu-west-1', aws_access_key_id=access_key, aws_secret_access_key=secret)
        self.logger = KhanLogger(origin=self.__class__.__name__)

    def modify_throughput(self, requested_read, requested_write, table):
        """
        Used to change the throughput of a specific table
        """
        read, write, num_dec_today, table_status = self.get_table_info(table)

        while requested_read != read or requested_write != write:

            self.logger.info(msg="Modifying {} from {}, {} to {}, {}".format(table.table_name, read, write,
                                                                             requested_read, requested_write))

            new_read, new_write = self._new_read_write(read, requested_read, write, requested_write)

            self.logger.info(msg="going to request read {} and write {}".format(new_read, new_write))

            if (new_read < read or new_write < write) and num_dec_today >= 4:
                # Todo - replace with custom error and handle in client code
                raise ValueError("Sorry, can't do any more decreases today.")
            table.update(throughput={'read': new_read, 'write': new_write})

            sleep_secs = 30
            table_status = 'UPDATING'
            self.logger.info(msg="Sleeping for {} secs before starting".format(sleep_secs))
            sleep(sleep_secs)
            while table_status == 'UPDATING':
                self.logger.info(msg="Sleeping for {} secs".format(sleep_secs))
                sleep(sleep_secs)
                read, write, num_dec_today, table_status = self.get_table_info(table)

        return read, write

    def _new_read_write(self, read, requested_read, write, requested_write):
        """
        Ensures that we change throughput in the correct amounts so as to not cause DDB to yell at us.
        """
        if requested_read == 0:
            read_change_prop = 0
        else:
            read_change_prop = requested_read / float(read)

        # max increase allowed is a doubling
        if read_change_prop > 2:
            new_read = read * 2
        else:
            new_read = requested_read

        if requested_write == 0:
            write_change_prop = 0
        else:
            write_change_prop = requested_write / float(write)

        if write_change_prop > 2:
            new_write = write * 2
        else:
            new_write = requested_write

        return new_read, new_write

    def get_table_info(self, table):
        """
        Returns meta information about the table, such as read speed, write speed, current status,
        and number of decreases today. Useful for figuring out how to change throughput.
        """
        desc = table.describe()
        status = desc['Table']['TableStatus']
        throughput = desc['Table']['ProvisionedThroughput']
        num_decreases = throughput['NumberOfDecreasesToday']
        read = throughput['ReadCapacityUnits']
        write = throughput['WriteCapacityUnits']
        return read, write, num_decreases, status




__author__ = 'mattjmorris'

from .dynamo import Dynamo
from boto.dynamodb2.table import Table
from pandas import DataFrame
from datetime import datetime
from khan_utils.encoding import to_unicode_or_bust, to_utf8_or_bust
from ut.coll import order_conserving


class DDBToBeSlurped(Dynamo):

    def __init__(self, access_key=None, secret=None):
        """
        ! Use test_mode factory method for instantiating this class with test_slurps and test_failed_slurps tables
        """
        super(DDBToBeSlurped, self).__init__(access_key, secret)

        self.table = Table('to_be_slurped', connection=self.connection)

    def save_info(self, search_terms):
        """
        search_terms can either be in the form of a list of dicts or else a single dict.
        If slurp_info is a list, batch write will be used
        """
        if isinstance(search_terms, str):
            search_terms = [search_terms]
        # search_terms = {'searchterm': search_terms}
        search_terms = [{'searchterm': x} for x in search_terms]
        # print search_terms
        with self.table.batch_write() as batch:
            for s in search_terms:
                batch.put_item(data=s, overwrite=True)

    def get_table(self, table_name=None):
        """
        Convenience method for client who may wish to get a specific table from the DynamoDB connection
        """
        table_name = table_name or self.table.table_name
        return Table(table_name, connection=self.connection)

    def truncate_table(self):
        """
        Delete whole table
        """
        with self.table.batch_write() as batch:
            for item in self.table.scan():
                batch.delete_item(searchterm=item['searchterm'])

    def modify_slurps_throughput(self, requested_read, requested_write):
        return self.modify_throughput(requested_read, requested_write, self.table)

    def get_slurps_table_info(self):
        return self.get_table_info(self.table)

    def get_table_as_df(self):
        return DataFrame([dict(r) for r in self.table.scan()])





"""
amazon_sender.py
~~~~~~~~

Python helper class that can send emails using Amazon SES and boto.
The biggest feature of this class is that encodings are handled properly.
It can send both text and html emails.
This implementation is using Python's standard library (which opens up for a lot more options).

Example::

    amazon_sender = AmazonSender(AWS_ID, AWS_SECRET)

    amazon_sender.send_email(sender=u'Me <john@doe.com>',
                             to='blah@blah.com',
                             subject='Hello friend',
                             text='Just a message',
                             html='<b>Just a message</b>',
                             sender_ascii='Ascii Sender <no_reply@wedoist.com>')


:copyright: 2011 by Amir Salihefendic ( http://amix.dk/ ).
:license: BSD
"""

import types

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE
from nippy.importing import get_environment_variable
from boto.ses import SESConnection


class AmazonSender(object):

    client = None

    def __init__(self, to_addresses=None, sender=None, aws_key=None, aws_secret=None):
        self.sender = sender or 'Thor Stats <thor_stats@mscoms.com>'
        self.to_addresses = to_addresses or 'thor@mscoms.com'
        self.aws_key = aws_key or get_environment_variable('VEN_S3_ACCESS_KEY')
        self.aws_secret = aws_secret or get_environment_variable('VEN_S3_SECRET')

    def send_email(self, subject='', text='', html=None, reply_addresses=None, sender_ascii=None):
        if not sender_ascii:
            sender_ascii = self.sender

        client = self.get_client()

        message = MIMEMultipart('alternative')
        message.set_charset('UTF-8')

        message['Subject'] = _encode_str(subject)
        message['From'] = _encode_str(self.sender)

        message['To'] = _convert_to_strings(self.to_addresses)

        if reply_addresses:
            message['Reply-To'] = _convert_to_strings(reply_addresses)

        message.attach(MIMEText(_encode_str(text), 'plain'))

        if html:
            message.attach(MIMEText(_encode_str(html), 'html'))

        return client.send_raw_email(message.as_string(),
                                     sender_ascii,
                                     destinations=self.to_addresses)

    def verify_email(self, email):
        client = self.get_client()
        return client.verify_email_address(email)

    def get_client(self):
        if not self.client:
            self.client = SESConnection(self.aws_key,
                                        self.aws_secret)
        return self.client


#--- Helpers ----------------------------------------------
def _convert_to_strings(list_of_strs):
    if isinstance(list_of_strs, (list, tuple)):
        result = COMMASPACE.join(list_of_strs)
    else:
        result = list_of_strs
    return _encode_str(result)

def _encode_str(s):
    if type(s) == str:
        return s.encode('utf8')
    return s


from boto.s3.connection import S3Connection
from boto.s3.key import Key
from boto.exception import S3CopyError, S3ResponseError
import tempfile
import pickle
import os
from nippy.importing import get_environment_variable
# import pygeonhole.utils as serialize_utils

# use case. I need to check all code that uses these methods, as well as the test notebook.


class S3(object):
    """
    Interaction with Amazon S3.
    """

    def __init__(self, bucket_name=None, base_folder=None, extension=None, force_extension=False, encoding=None,
                 access_key=None, secret=None):
        """
        Creates an instance with a handle on the S3 bucket corresponding to bucket_name.

        If access_key and/or secret are not passed in, assumes we are accessing erenev's aws account and that the
        access info is stored as environment variables on the current server.

        Connection and bucket are available to clients via self properties, in case clients wish to use those objects
        directly.
        """
        assert not extension, "extension has not been implement yet for S3."

        if access_key and not secret:
            if access_key == 'ut':
                access_key = get_environment_variable('VEN_AWS_ACCESS_KEY_ID')
                secret = get_environment_variable('VEN_AWS_SECRET_ACCESS_KEY')
            elif access_key == 'mon':
                access_key = get_environment_variable('MON_AWS_ACCESS_KEY_ID')
                secret = get_environment_variable('MON_AWS_SECRET_ACCESS_KEY')
            else:
                ValueError('I cannot recognize that access_key')
        else:  # if access_key is not given, take a default
            #access_key = access_key or os.environ['MON_AWS_ACCESS_KEY_ID']
            #secret = secret or os.environ['MON_AWS_SECRET_ACCESS_KEY']
            access_key = access_key or get_environment_variable('VEN_AWS_ACCESS_KEY_ID')
            secret = secret or get_environment_variable('VEN_AWS_SECRET_ACCESS_KEY')

        # note - this calls the setter
        self.base_folder = base_folder
        self.extension = extension
        self.force_extension = force_extension
        self.encoding = encoding

        self.connection = S3Connection(access_key, secret, host='s3-eu-west-1.amazonaws.com')
        if bucket_name:
            self.bucket = self.connection.get_bucket(bucket_name)
        else:
            self.bucket = None


    @property
    def base_folder(self):
        return self._base_folder or ''

    @base_folder.setter
    def base_folder(self, base_folder):
        self._base_folder = self._ensure_good_folder_name(base_folder)

    def dumpo(self, obj, key_name, folder=None, bucket_name=''):
        """
        --
        For saving objects to S3
        Do not use this method for saving strings. It will work, but using dumps() is more efficient.
        --
        params:
          obj: the object to save
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket_name: optional - overrides the default bucket name passed in during init.
        return:
          number of bytes
        """
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_dump(key_name, folder, bucket)

        with tempfile.TemporaryFile() as tempf:
            pickle.dump(obj, tempf)
            tempf.seek(0)
            return s3_key.set_contents_from_file(tempf)

    def dumpf(self, f, key_name, folder=None, bucket_name=''):
        """
        --
        For saving objects already in the form of a file (usually tempfile) to S3
        Note that it is the responsibility of the client to manage the file, including setting seek(0) if necessary
        before passing it here, and then closing the file if needed afterwards.
        --
        params:
          f: the file to save
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket_name: optional - overrides the default bucket name passed in during init.
        return:
          number of bytes
        """
        if isinstance(f, str) and os.path.exists(f):
            f = open(f, 'r')
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_dump(key_name, folder, bucket)
        return s3_key.set_contents_from_file(f)

    def dumps(self, the_str, key_name, folder=None, bucket_name=None):
        """
        --
        For saving strings to S3
        --
        params:
          the_str: the string to save
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket: optional - overrides the default bucket name passed in during init.
        return:
          number of bytes
        """
        assert isinstance(the_str,
                          str), 'the_str must be an instance of basestring, but was an instance of {}'.format(type(the_str))
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_dump(key_name, folder, bucket)
        return s3_key.set_contents_from_string(the_str)

    def loado(self, key_name, folder=None, bucket_name='', local_file_name=None, deserialize_f=lambda x: pickle.load(x)):
        """
        --
        For loading objects from S3
        Do not use this method for loading strings since de-pickling will throw an error.
        --
        params:
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket: optional - overrides the default bucket name passed in during init.
        return:
          the serialized object
        """
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_load(key_name, folder, bucket)
        # if client passed in a local file name, download to that file, else just return the contents
        if local_file_name:
            s3_key.get_contents_to_filename(local_file_name)
        else:
            with tempfile.TemporaryFile() as tempf:
                s3_key.get_contents_to_file(tempf)
                tempf.seek(0)
                val = deserialize_f(tempf)
                return val

    def loads(self, key_name, folder=None, bucket_name=None, local_file_name=None):
        """
        --
        For loading strings from S3
        --
        params:
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket: optional - overrides the default bucket name passed in during init.
        return:
          the serialized string
        """
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_load(key_name, folder, bucket)
        # if client passed in a local file name, download to that file, else just return the contents
        if local_file_name:
            s3_key.get_contents_to_filename(local_file_name)
        else:
            return s3_key.get_contents_as_string()

    def loadf(self, key_name, local_file_name, folder=None, bucket_name=None):
        """
        --
        For downloading files from S3
        --
        params:
          key_name: the value that will be assigned to key.key. May be full path, including folders.
          folder: optional - the folder name(s). Only use this param if you did not use full path in key_name
          bucket: optional - overrides the default bucket name passed in during init.
        return:
          the serialized string
        """
        bucket = self._get_new_bucket_or_default(bucket_name)
        s3_key = self._get_s3_key_for_load(key_name, folder, bucket)
        # if client passed in a local file name, download to that file, else just return the contents
        s3_key.get_contents_to_filename(local_file_name)


    def mk_key_name(self, key_name, folder=None, bucket=None, **kwargs):
        # have bucket_name overwrite existing bucket
        if 'bucket_name' in list(kwargs.keys()):
            bucket = self.connection.get_bucket(kwargs['bucket_name'])
        else:
            bucket = bucket or self.bucket
        return os.path.join(folder or self.base_folder or '', key_name)

    def get_key(self, key_name, folder=None, bucket=None, **kwargs):
        """
        Takes care of the process of getting an s3 key
        """
        if isinstance(key_name, str):
            if 'bucket_name' in list(kwargs.keys()):
                bucket = self.connection.get_bucket(kwargs['bucket_name'])
            else:
                bucket = bucket or self.bucket
            key_full_name = self.mk_key_name(key_name, folder=folder, bucket=bucket, **kwargs)
            key = bucket.lookup(key_full_name)
            if not key:
                raise MissingS3KeyError("%s not found in %s" % (key_full_name, bucket.name))
            return key
        elif isinstance(key_name, Key):
            return key_name  # it's actually a boto.s3.key.Key already


    def get_http_for_key(self, key_name, folder=None, bucket=None,
                         expires_in=15*24*60*60, query_auth=True, force_http=True, **kwargs):
        key = self.get_key(key_name, folder=folder, bucket=bucket, **kwargs)
        return key.generate_url(expires_in=expires_in, query_auth=query_auth, force_http=force_http)


    def get_https_for_key(self, key_name, folder=None, bucket=None, expires_in=4*24*60*60, **kwargs):
        key = self.get_key(key_name, folder=folder, bucket=bucket, **kwargs)
        return key.generate_url(expires_in=expires_in, query_auth=True, force_http=False)


    def get_all_keys(self, folder, clean=True, bucket_name=''):
        """
        Returns a generator of all keys found in the specified folder and (optional=default) bucket.
        If clean = True (default), the empty 'folder' key that is returned by S3 API is removed and
        the key names have the folder name removed from them.
        If clean = False, the results are returned as-is from bucket.list for the folder.
        """
        bucket = self._get_new_bucket_or_default(bucket_name)

        if not folder.endswith('/'):
            folder += '/'

        key_result_set = bucket.list(prefix=folder)

        if clean:
            key_result_set_no_empty = (k for k in key_result_set if k.name.replace(folder, '') != '')
            key_result_set_no_folder_names = (self._remove_folder_from_name(k, folder) for k in key_result_set_no_empty)
            return key_result_set_no_folder_names
        else:
            return key_result_set

    def update_metadata(self, metadata_dict, key_name=None, folder='', key=None, bucket_name=''):
        """
        In order to update metadata on an object, need to copy (==resave) it to same location
        Note: pass in EITHER an actual key or else a key name (for lookup)
        """
        bucket = self._get_new_bucket_or_default(bucket_name)
        key = key or bucket_name.lookup(self._full_key_name(folder, key_name))
        key.metadata.update(metadata_dict)
        # MJM - note, removing preserve ACL since we should not need it here
        key.copy(bucket.name, key.name, key.metadata)

    def copy_and_return_errors(self, key_name_list, from_folder='', from_bucket_name='',
                               to_folder='', to_bucket_name='', storage_class='REDUCED_REDUNDANCY'):
        """
        Take a list of key names and moves them from the specified bucket and folder to the specified bucket and folder.
        If no folder(s) is specified, the top level of the bucket is used.
        If no bucket(s) is specified, the default bucket is used
        Returns a list of all errors encountered
        """

        assert storage_class in ['STANDARD', 'REDUCED_REDUNDANCY']

        from_bucket = self._get_new_bucket_or_default(from_bucket_name)
        to_bucket = bucket = self._get_new_bucket_or_default(to_bucket_name)

        to_and_from_key_names = [
            (self._full_key_name(to_folder, key_name),
             self._full_key_name(from_folder, key_name))
            for key_name in key_name_list
        ]

        errors = []

        for to_from_names in to_and_from_key_names:
            try:
                to_bucket.copy_key(
                    new_key_name=to_from_names[0],
                    src_bucket_name=from_bucket.name,
                    src_key_name=to_from_names[1],
                    storage_class=storage_class
                )
            except S3ResponseError as e:
                errors.append(e)

        return errors

    def _remove_folder_from_name(self, key, folder):
        key.name = key.name.replace(folder, '')
        return key

    def _get_s3_key_for_dump(self, key_name, folder, bucket):
        """
        Takes care of the process of getting an s3 key object set up for saving, within the correct bucket and
        with a name that takes the folder (if any) into account.
        """
        s3_key = Key(bucket)
        s3_key.key = self._full_key_name(folder, key_name)
        return s3_key

    def _get_s3_key_for_load(self, key_name, folder, bucket):
        """
        Takes care of the process of getting an s3 key object set up for loading, within the correct bucket and
        with a name that takes the folder (if any) into account.
        """
        key = bucket.lookup(self._full_key_name(folder, key_name))
        if not key:
            raise MissingS3KeyError("%s not found in %s" % (key_name, bucket.name))
        return key


    def _full_key_name(self, folder, key_name):
        folder_path = self.base_folder + self._ensure_good_folder_name(folder)
        return folder_path + key_name

    def _ensure_good_folder_name(self, folder_name):

        # we are going to append this to key name, so ensure it is not null
        folder_name = folder_name or ''

        if folder_name:
            if not folder_name.endswith('/'):
                folder_name += '/'
        return folder_name

    def _get_new_bucket_or_default(self, bucket_name):
        if bucket_name:
            return self.connection.get_bucket(bucket_name)
        else:
            return self.bucket


class MissingS3KeyError(Exception):
    pass



__author__ = 'mattjmorris'

from .dynamo import Dynamo
from boto.dynamodb2.table import Table
from datetime import datetime, date, timedelta
from pandas import DataFrame
import re

DAY_STR_FORMAT = "%Y-%m-%d"
DAY_STR_RE = re.compile(r'^(\d{4})-(\d{2})-(\d{2})$')
SECOND_STR_FORMAT = "%Y-%m-%d %H:%M:%S"
SECOND_STR_RE = re.compile(r'^(\d{4})-(\d{2})-(\d{2})\s(\d{2}:\d{2}:\d{2})$')

class DDBRuns(Dynamo):

    @classmethod
    def from_test_mode(cls, access_key=None, secret=None):
        """
        Use this for getting an instance of this class that uses test tables.
        """
        instance = cls(access_key, secret)
        instance.table = Table('test_runs', connection=instance.connection)
        return instance

    def __init__(self, access_key=None, secret=None):
        """
        When called directly (as should be done for production code), sets table to the production 'runs' table.
        """
        super(DDBRuns, self).__init__(access_key, secret)
        self.table = Table('runs', connection=self.connection)

    def save_new_run(self, dt_str=None, start_date_str=None, end_date_str=None):
        """
        dt_str = datetime of run. Defaults to now.
        start_date_str = the start date for look-back of query performance data processing. * No default
        end_date_str = the end date for query performance data processing. Defaults to today.
        """
        assert start_date_str, "start_date_str is required when saving a new run to runs table."
        assert DAY_STR_RE.match(start_date_str)
        if end_date_str:
            assert DAY_STR_RE.match(end_date_str)
        if dt_str:
            assert SECOND_STR_RE.match(dt_str)

        dt_str = dt_str or datetime.now().strftime(SECOND_STR_FORMAT)
        end_date_str = end_date_str or datetime.now().strftime(DAY_STR_FORMAT)
        return self.table.put_item(data={'dt': dt_str, 'start': start_date_str, 'end': end_date_str})

    def most_recent_start_date_str(self):
        """
        :return: a string representing most recent start date from db
        """
        df = self.get_runs_df()
        if df.empty:
            return None
        else:
            # should already be sorted, but just in case...
            df.sort(columns=['dt'], ascending=True, inplace=True)
            return df.iloc[len(df)-1]['start']

    def most_recent_end_date_str(self):
        """
        :return: a string representing most recent end date from db
        """
        df = self.get_runs_df()
        if df.empty:
            return None
        else:
            # should already be sorted, but just in case...
            df.sort(columns=['dt'], ascending=True, inplace=True)
            return df.iloc[len(df)-1]['end']

    def get_runs_df(self):
        """
        Returns all table as dataframe, sorted with most recent entry on bottom (ascending order)
        """
        df = DataFrame([{k: v for k, v in list(r.items())} for r in self.table.scan()])
        if df.empty:
            return df
        else:
            df.sort(columns=['dt'], ascending=True, inplace=True)
            # force df to have columns in this order
            return df[['dt', 'start', 'end']]

    def modify_throughput(self, requested_read, requested_write, table=None):
        table = table or self.table
        return super(DDBRuns, self).modify_throughput(requested_read, requested_write, table)

    def truncate_table(self):
        """
        WARNING! Only use for test mode table
        """
        assert self.table.table_name == 'test_runs', "Will only truncate test table. To truncate production table, run code manually"
        with self.table.batch_write() as batch:
            for item in self.table.scan():
                batch.delete_item(dt=item['dt'])


    def thors_start_end_date_strings(self, new_run=True, days_ago_start=30):
        if new_run:
            if days_ago_start is not None:
                print(days_ago_start)
                start_date_str = self._days_ago_str(days_ago_start)
            else:
                start_date_str = self.most_recent_end_date_str()
            end_date_str = date.today().strftime(DAY_STR_FORMAT)
        else:
            start_date_str = self.most_recent_start_date_str()
            end_date_str = self.most_recent_end_date_str()
            assert start_date_str, "Start date string is None, please check the database since we are not doing a new run"
            assert end_date_str, "End date string is None, please check the database since we are not doing a new run"
        return start_date_str, end_date_str

    def _days_ago_str(self, num_days_ago):
        return (date.today() - timedelta(days=num_days_ago)).strftime(DAY_STR_FORMAT)

    def start_end_date_strings(self, new_run=True, days_ago_start=30):
        if new_run:
            start_date_str = self.most_recent_end_date_str() or self._days_ago_str(days_ago_start)
            end_date_str = date.today().strftime(DAY_STR_FORMAT)
        else:
            start_date_str = self.most_recent_start_date_str()
            end_date_str = self.most_recent_end_date_str()
        return start_date_str, end_date_str
